"""File utility tools"""
import re
from typing import Mapping


def replace_in_files(files_map: Mapping[str, str], replace_map: Mapping[str, str]):
    """
    Replaces a series of matches in a collection of files

    Args:
        file_paths (dict): A dictionary where the key is the file source and the value is a file target.
            If the target is set to an empty string, the original file will be replaced
        replace_map (dict): A dictionary where the keys are the regex to be matched and
            the value is the content to replace the match with

    Example:

        >>> file_map = {'/tmp/source.txt': '/tmp/destination.txt'}
        >>> replace_map = {'.*foo.*': 'bar'}
        >>> replace_in_files(file_map, replace_map)

    """
    for file_src_path, file_dst_path in files_map.items():
        with open(file_src_path, "r+") as src_file:
            data = src_file.read()
            new_data = data
            for pattern, replace_value in replace_map.items():
                new_data = re.sub(pattern, replace_value, new_data, flags=re.DOTALL)
            if not file_dst_path:
                src_file.seek(0)
                src_file.write(new_data)
                src_file.truncate()
            else:
                with open(file_dst_path, "w") as dest_file:
                    dest_file.write(new_data)



"""Directory utility tools"""
import os
from typing import Iterator


def list_files(path: str, recursive: bool = False) -> Iterator[str]:
    """List of the absolute path of all files within a folder.

    Args:
        path (str): The folder to be scanned for files.
        recursive (bool, optional): Defaults to False. Whether to include all subdirectories.

    Yields:
        str: The next file's absolute path within the folder
    """

    if os.path.isdir(path):
        for dirpath, dirnames, filenames in os.walk(path):
            for _file in filenames:
                yield os.path.abspath(os.path.join(dirpath, _file))
            if not recursive:
                break



"""Dictionary utility tools"""
from typing import Any


def merge(*dicts: dict) -> dict:
    """Deep merges the right most dictionaries into the left most one.

    Args:
        *dicts (dict): dictionaries to be merged.

    Examples:

        >>> a = {'a': 1, 'b': 2}
        >>> b = {'b': 3, 'c': {'c1': 4, 'c2': 3}}
        >>> c = {'a': 3, 'c': {'c1': 3 }}
        >>> print(merge(a,b,c))
        {'c': {'c1': 3, 'c2': 3}, 'a': 3, 'b': 3}

    Returns:
        dict
    """

    def _merge(dict1, dict2):
        for k in set(dict1.keys()).union(dict2.keys()):
            if k in dict1 and k in dict2:
                if isinstance(dict1[k], dict) and isinstance(dict2[k], dict):
                    yield (k, dict(merge(dict1[k], dict2[k])))
                else:
                    # If one of the values is not a dict, you can't continue merging it.
                    # Value from second dict overrides one in first and we move on.
                    yield (k, dict2[k])
            elif k in dict1:
                yield (k, dict1[k])
            else:
                yield (k, dict2[k])

    result = dict()
    for index, _dict in reversed(list(enumerate(dicts))):
        if not result:
            result = dict(_merge(dicts[index - 1], _dict))
        result = dict(_merge(dicts[index - 1], result))

    return dict(result)


def get_nested(the_dict: dict, nested_key: str, default: Any = None) -> Any:
    """
    Returns the value of a nested key in an dictorinary

    Args:
        the_dict (dict): The dictionary to get the nested value from
        nested_key (str): The nested key
        default: The value to be returned if the key is not found

    Returns:

    Examples:
        >>> t = {'a': {'b': {'c':'value_of_c'}}}
        >>> get_nested(t, 'b:c')
        'value_of_c'
        >>> get_nested(t, 'b:c:d', 'not found')
        'not found'

    """

    keys = nested_key.split(":")

    if not hasattr(the_dict, "__iter__"):
        return default

    if len(keys) == 1:
        if keys[0] in the_dict:
            return the_dict[keys[0]]
        else:
            return default

    if keys[0] in the_dict:
        return get_nested(the_dict[keys[0]], ":".join(keys[1:]), default)
    else:
        return default


def update_nested(the_dict: dict, nested_key: str, value: Any):
    """
    Updates nested keys inside a dictionary

    Args:
        the_dict (dict): The dictionary to be updated
        nested_key (str): The nested key
        value: The value to be set to the nested key

    Examples:
        >>> t = {}
        >>> update_nested(t, 'a:b:c', 1)
        >>> t
        {'a': {'b': {'c': 1}}}
        >>> update_nested(t, 'a:b:d', 1)
        >>> t
        {'a': {'b': {'c': 1, 'd': 1}}}
        >>> update_nested(t, 'a:b:c', 2)
        >>> t
        {'a': {'b': {'c': 2, 'd': 1}}}
    """
    keys = nested_key.split(":")

    if len(keys) == 1:
        the_dict.update({keys[0]: value})
    else:
        if keys[0] not in the_dict:
            the_dict[keys[0]] = {}
        update_nested(the_dict[keys[0]], ":".join(keys[1:]), value)

    return the_dict



"""YAML utility tools"""
import os
from re import match

import yaml
from jinja2 import Template

from codebox import dir_tools
from codebox import dict_tools


def load(
    path: str,
    recursive: bool = False,
    match_pattern: str = None,
    ignore_empty: bool = False,
    parse_jinja: bool = False,
    jinja_context: dict = None,
) -> dict:
    """Loads and parses a file or a folder containing yaml files.
    All existing dictionaries will be deeply merged.
    Merge order respects file path sorted by name.

    Args:
        path (str): File or directory path.
        recursive (bool, optional): Defaults to False. Whether or not to include subdirectoties.
        match_pattern (str, optional): A regular expression to be used to filter the files to be load
            based on the file's full name.
        ignore_empty (bool, optional): Whether to ignore empty files.
        parse_jinja (bool, optional): Whether to parse Jinja code.
        jinja_context (dict, optional): The Jinja Context.

    Returns:
        dict

    """

    files = []
    if os.path.isdir(path):
        files = sorted(dir_tools.list_files(path, recursive))
    else:
        files = [path]

    data = dict()
    for _file in files:
        if match_pattern and not match(match_pattern, _file):
            continue
        with open(_file) as fobj:
            try:
                file_content = fobj.read()
                if parse_jinja:
                    template = Template(file_content)
                    file_content = template.render(jinja_context or dict())

                data = dict_tools.merge(data, yaml.load(file_content, Loader=yaml.FullLoader))

            except AttributeError:
                if not ignore_empty:
                    raise

    return data



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['codebox']

package_data = \
{'': ['*']}

install_requires = \
['Jinja2>=3.1.2,<4.0.0', 'PyYAML>=6.0,<7.0']

extras_require = \
{':extra == "docs"': ['sphinx-rtd-theme>=1.0.0,<2.0.0',
                      'sphinxcontrib-napoleon>=0.7,<0.8']}

setup_kwargs = {
    'name': 'codebox',
    'version': '1.1.0',
    'description': 'A collection of Python utility modules',
    'long_description': 'codebox - Python tools collection.\n-----------------------------------------\n\nInstallation\n==================\n\n.. code-block:: bash\n\n    pip install codebox\n\n\nDocumentation\n=============\n\nhttps://wils0ns.github.io/codebox/\n',
    'author': 'Wilson Santos',
    'author_email': 'wilson@codeminus.org',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/wils0ns/codebox',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
